import typing as t
from inspect import isawaitable

from hiku.federation.version import DEFAULT_FEDERATION_VERSION

from hiku.engine import pass_context

from hiku.directives import SchemaDirective
from hiku.enum import BaseEnum
from hiku.federation.utils import get_entity_types
from hiku.scalar import Scalar
from hiku.federation.scalars import _Any, FieldSet, LinkImport

from hiku.types import Optional, Record, Sequence, String, TypeRef, UnionRef

from hiku.graph import (
    Field,
    Graph as _Graph,
    GraphTransformer,
    Interface,
    Link,
    Node,
    Option,
    Root,
    Union,
)


class FederatedNode(Node):
    def __init__(
        self,
        name: t.Optional[str],
        fields: t.List[t.Union[Field, Link]],
        *,
        description: t.Optional[str] = None,
        directives: t.Optional[t.Sequence[SchemaDirective]] = None,
        resolve_reference: t.Optional[t.Callable] = None,
    ):
        super().__init__(
            name, fields, description=description, directives=directives
        )
        self.resolve_reference = resolve_reference


class GraphInit(GraphTransformer):
    type_to_resolve_reference_map: t.Dict[str, t.Callable]
    is_async: bool
    has_entity_types: bool

    @classmethod
    def init(
        cls,
        items: t.List[t.Union[FederatedNode, Node]],
        is_async: bool,
        has_entity_types: bool = False,
    ) -> t.List[Node]:
        self = cls()
        self.type_to_resolve_reference_map = {}
        self.is_async = is_async
        self.has_entity_types = has_entity_types
        return [self.visit(i) for i in items]

    def visit_root(self, root: Root) -> Root:
        if not self.has_entity_types:
            return root

        fields = root.fields[:]
        if "_entities" not in root.fields_map:
            fields.append(self.entities_link())
        if "_service" not in root.fields_map:
            fields.append(self.service_field())

        return Root(fields)

    def visit_node(self, obj: Node) -> Node:
        if hasattr(obj, "resolve_reference") and obj.name is not None:
            self.type_to_resolve_reference_map[obj.name] = obj.resolve_reference

        return super(GraphInit, self).visit_node(obj)

    def entities_link(self) -> Link:
        def entities_resolver(
            options: t.Dict,
        ) -> t.List[t.Tuple[t.Any, t.Type[TypeRef]]]:
            representations = options["representations"]
            if not representations:
                return []

            typ = representations[0]["__typename"]

            if typ not in self.type_to_resolve_reference_map:
                raise TypeError(
                    'Type "{}" must have "reference_resolver"'.format(typ)
                )

            resolve_reference = self.type_to_resolve_reference_map[typ]
            return [
                (r, TypeRef[typ]) for r in resolve_reference(representations)
            ]

        def _asyncify(func: t.Callable) -> t.Callable:
            async def wrapper(*args: t.Any, **kwargs: t.Any) -> t.List[t.Tuple]:
                result = []
                for item in func(*args, **kwargs):
                    if isawaitable(item[0]):
                        item[0] = await item[0]
                    result.append(item)
                return result

            return wrapper

        return Link(
            "_entities",
            Sequence[Optional[UnionRef["_Entity"]]],
            _asyncify(entities_resolver)
            if self.is_async
            else entities_resolver,
            options=[
                Option("representations", Sequence[_Any]),
            ],
            requires=None,
        )

    def service_field(self) -> Field:
        @pass_context
        def service_resolver(
            ctx: t.Dict,
            fields: t.List,
        ) -> t.List:
            return [{"sdl": ctx["__sdl__"]}]

        def _async(func: t.Callable) -> t.Callable:
            @pass_context
            async def wrapper(*args: t.Any, **kwargs: t.Any) -> t.List[t.Dict]:
                return func(*args, **kwargs)

            return wrapper

        return Field(
            "_service",
            TypeRef["_Service"],
            _async(service_resolver) if self.is_async else service_resolver,  # type: ignore[arg-type]  # noqa: E501
        )


class Graph(_Graph):
    def __init__(
        self,
        items: t.List[t.Union[FederatedNode, Node]],
        data_types: t.Optional[t.Dict[str, t.Type[Record]]] = None,
        directives: t.Optional[t.Sequence[t.Type[SchemaDirective]]] = None,
        unions: t.Optional[t.List[Union]] = None,
        interfaces: t.Optional[t.List[Interface]] = None,
        enums: t.Optional[t.List[BaseEnum]] = None,
        scalars: t.Optional[t.List[t.Type[Scalar]]] = None,
        is_async: bool = False,
    ):
        self.is_async = is_async

        if unions is None:
            unions = []

        if scalars is None:
            scalars = []

        unions_map = {union.name: union for union in unions}
        scalars_map = {scalar.__type_name__: scalar for scalar in scalars}

        entity_types = get_entity_types(items, DEFAULT_FEDERATION_VERSION)
        if entity_types:
            if "_Entity" not in unions_map:
                unions.append(Union("_Entity", entity_types))

        for scalar in [_Any, FieldSet, LinkImport]:
            if scalar.__type_name__ not in scalars_map:
                scalars.append(scalar)

        if data_types is None:
            data_types = {}

        if "_Service" not in data_types:
            data_types["_Service"] = Record[{"sdl": String}]

        items = GraphInit.init(items, is_async, bool(entity_types))

        super().__init__(
            items, data_types, directives, unions, interfaces, enums, scalars
        )

    @classmethod
    def from_graph(
        cls,
        other: "Graph",
        root: Root,
    ) -> "Graph":
        """Create graph from other graph, with new root node.
        Useful for creating mutation graph from query graph.
        """
        return cls(
            other.nodes + [root],
            data_types=other.data_types,
            directives=other.directives,
            unions=other.unions,
            interfaces=other.interfaces,
            enums=other.enums,
            scalars=other.scalars,
            is_async=other.is_async,
        )



import typing as t

from contextlib import contextmanager
from collections import abc as collections_abc

from hiku.scalar import ScalarMeta

from ..types import (
    AbstractTypeVisitor,
    IDMeta,
    Record,
    OptionalMeta,
    SequenceMeta,
    RecordMeta,
    TypeRefMeta,
    EnumRefMeta,
    GenericMeta,
    AnyMeta,
    BooleanMeta,
    StringMeta,
    IntegerMeta,
    FloatMeta,
    MappingMeta,
)

from hiku.query import (
    Field as QueryField,
    FieldBase,
    Fragment,
    Node as QueryNode,
    Link as QueryLink,
    QueryVisitor,
)
from hiku.graph import (
    Interface,
    LinkType,
    Node,
    Field,
    Link,
    GraphVisitor,
    Root,
    Option,
    Nothing,
    Graph,
    Union,
)
from .errors import Errors


_undefined = object()


OptionValue = t.Union[
    int, str, float, bool, collections_abc.Sequence, collections_abc.Mapping
]


class _AssumeRecord(AbstractTypeVisitor):
    def __init__(
        self, data_types: t.Dict[str, t.Type[Record]], _nested: bool = False
    ):
        self._data_types = data_types
        self._nested = _nested

    def _get_nested(self) -> "_AssumeRecord":
        return _AssumeRecord(self._data_types, _nested=True)

    def visit(self, obj: t.Any) -> t.Any:
        if obj is not None:
            return obj.accept(self)

    def _false(self, obj: t.Any) -> None:
        pass

    visit_any = _false
    visit_boolean = _false
    visit_string = _false
    visit_id = _false
    visit_integer = _false
    visit_float = _false
    visit_mapping = _false
    visit_callable = _false
    visit_unionref = _false
    visit_interfaceref = _false
    visit_enumref = _false
    visit_scalar = _false

    def visit_optional(self, obj: OptionalMeta) -> t.Optional[t.OrderedDict]:
        if not self._nested:
            return self._get_nested().visit(obj.__type__)
        return None

    def visit_sequence(self, obj: SequenceMeta) -> t.Optional[t.OrderedDict]:
        if not self._nested:
            return self._get_nested().visit(obj.__item_type__)
        return None

    def visit_record(self, obj: RecordMeta) -> t.OrderedDict:
        # return fields alongside type definitions
        return obj.__field_types__

    def visit_typeref(self, obj: TypeRefMeta) -> t.OrderedDict:
        return self.visit(self._data_types[obj.__type_name__])


class _AssumeField(GraphVisitor):
    def __init__(self, node: Node, errors: Errors) -> None:
        self.node = node
        self.errors = errors

    def visit_field(self, obj: Field) -> bool:
        return True

    def visit_link(self, obj: Link) -> bool:
        self.errors.report(
            'Trying to query "{}.{}" link as it was a field'.format(
                self.node.name or "root", obj.name
            )
        )
        return False

    def visit_node(self, obj: Node) -> bool:
        assert (
            self.node.name is None
        ), "Nested node can be only in the root node"
        self.errors.report(
            'Trying to query "{}" node as it was a field'.format(obj.name)
        )
        return False

    def visit_root(self, obj: Root) -> t.NoReturn:
        raise AssertionError("Root node is not expected here")


class _OptionError(TypeError):
    def __init__(self, description: str) -> None:
        self.description = description
        super(_OptionError, self).__init__(description)


class _OptionTypeError(_OptionError):
    def __init__(self, value: OptionValue, expected: GenericMeta) -> None:
        description = '"{}" instead of {!r}'.format(
            type(value).__name__, expected
        )
        super(_OptionTypeError, self).__init__(description)


class _OptionTypeValidator:
    def __init__(
        self, data_types: t.Dict[str, t.Type[Record]], value: OptionValue
    ) -> None:
        self._data_types = data_types
        self._value = [value]

    @property
    def value(self) -> OptionValue:
        return self._value[-1]

    @contextmanager
    def push(self, value: OptionValue) -> t.Iterator[None]:
        self._value.append(value)
        try:
            yield
        finally:
            self._value.pop()

    def visit(self, type_: GenericMeta) -> None:
        type_.accept(self)  # type: ignore

    def visit_any(self, type_: AnyMeta) -> None:
        pass

    def visit_boolean(self, type_: BooleanMeta) -> None:
        if not isinstance(self.value, bool):
            raise _OptionTypeError(self.value, type_)
        return None

    def visit_string(self, type_: StringMeta) -> None:
        if not isinstance(self.value, str):
            raise _OptionTypeError(self.value, type_)
        return None

    def visit_id(self, type_: IDMeta) -> None:
        if not isinstance(self.value, str):
            raise _OptionTypeError(self.value, type_)
        return None

    def visit_integer(self, type_: IntegerMeta) -> None:
        if not isinstance(self.value, int):
            raise _OptionTypeError(self.value, type_)
        return None

    def visit_float(self, type_: FloatMeta) -> None:
        if not isinstance(self.value, float):
            raise _OptionTypeError(self.value, type_)
        return None

    def visit_optional(self, type_: OptionalMeta) -> None:
        if self.value is not None:
            self.visit(type_.__type__)
        return None

    def visit_sequence(self, type_: SequenceMeta) -> None:
        if not isinstance(self.value, collections_abc.Sequence):
            raise _OptionTypeError(self.value, type_)
        for item in self.value:
            with self.push(item):
                self.visit(type_.__item_type__)

        return None

    def visit_mapping(self, type_: MappingMeta) -> None:
        if not isinstance(self.value, collections_abc.Mapping):
            raise _OptionTypeError(self.value, type_)
        for key, value in self.value.items():
            with self.push(key):
                self.visit(type_.__key_type__)
            with self.push(value):
                self.visit(type_.__value_type__)

        return None

    def visit_record(self, type_: RecordMeta) -> None:
        if not isinstance(self.value, collections_abc.Mapping):
            raise _OptionTypeError(self.value, type_)

        unknown = set(self.value).difference(type_.__field_types__)
        if unknown:
            fields = ", ".join(sorted(map(repr, unknown)))
            raise _OptionError("unknown fields: {}".format(fields))

        missing = set(type_.__field_types__).difference(self.value)
        if missing:
            fields = ", ".join(sorted(missing))
            raise _OptionError("missing fields: {}".format(fields))

        for key, value_type in type_.__field_types__.items():
            with self.push(self.value[key]):
                self.visit(value_type)

        return None

    def visit_typeref(self, type_: TypeRefMeta) -> None:
        assert (
            type_.__type_name__ in self._data_types
        ), f'"{type_.__type_name__}" type is not present in graph data_types'
        self.visit(self._data_types[type_.__type_name__])

    # TODO: add scalar validation errors
    def visit_scalar(self, type_: ScalarMeta) -> None:
        pass

    # TODO: add enum validation errors
    def visit_enumref(self, type_: EnumRefMeta) -> None:
        pass


class _ValidateOptions(GraphVisitor):
    """
    Validate options for fields and links.

    :param data_types: Mapping of data types.
    :param options: Options to validate.
    :param for_: Path to the field or link. Used in error messages.
    :param errors: Errors container.
    """

    def __init__(
        self,
        data_types: t.Dict[str, t.Type[Record]],
        options: t.Optional[t.Dict],
        for_: t.Tuple[t.Any, ...],
        errors: Errors,
    ) -> None:
        self._data_types = data_types
        self.options = options
        self.for_ = for_
        self.errors = errors
        self._options = options or {}

    def visit_link(self, obj: Link) -> None:
        super(_ValidateOptions, self).visit_link(obj)
        unknown = set(self._options).difference(obj.options_map)
        if unknown:
            node, field = self.for_
            self.errors.report(
                'Unknown options for "{}.{}": {}'.format(
                    node, field, ", ".join(unknown)
                )
            )

        return None

    visit_field = visit_link  # type: ignore

    def visit_option(self, obj: Option) -> None:
        value = self._options.get(obj.name, obj.default)
        if value is Nothing:
            node, field = self.for_
            self.errors.report(
                'Required option "{}.{}:{}" is not specified'.format(
                    node, field, obj.name
                )
            )
        elif obj.type is not None:
            try:
                _OptionTypeValidator(self._data_types, value).visit(obj.type)
            except _OptionError as err:
                node, field = self.for_
                self.errors.report(
                    'Invalid value for option "{}.{}:{}", {}'.format(
                        node, field, obj.name, err.description
                    )
                )

        return None

    def visit_node(self, obj: Node) -> None:
        assert self.options is None, "Node can not have options"

    def visit_root(self, obj: Root) -> t.NoReturn:
        raise AssertionError("Root node is not expected here")


class _RecordFieldsValidator(QueryVisitor):
    def __init__(
        self,
        data_types: t.Dict[str, t.Type[Record]],
        field_types: t.OrderedDict,
        errors: Errors,
    ) -> None:
        self._data_types = data_types
        self._field_types = field_types
        self._errors = errors

    def visit_field(self, obj: QueryField) -> None:
        if obj.name == "__typename":
            return

        if obj.name not in self._field_types:
            self._errors.report('Unknown field name "{}"'.format(obj.name))
        elif obj.options is not None:
            self._errors.report("Options are not expected")
        elif _AssumeRecord(self._data_types).visit(self._field_types[obj.name]):
            self._errors.report(
                'Trying to query "{}" link as it was a field'.format(obj.name)
            )

    def visit_link(self, obj: QueryLink) -> None:
        field_types = _AssumeRecord(self._data_types).visit(
            self._field_types[obj.name]
        )
        if field_types is not None:
            fields_validator = _RecordFieldsValidator(
                self._data_types, field_types, self._errors
            )
            for field in obj.node.fields:
                fields_validator.visit(field)
        else:
            self._errors.report('"{}" is not a link'.format(obj.name))

    def visit_node(self, obj: QueryNode) -> None:
        raise AssertionError("Node is not expected here")


def _field_eq(a: FieldBase, b: FieldBase) -> bool:
    return a.name == b.name and a.options == b.options


class DefaultQueryValidator(QueryVisitor):
    """
    Validate query against graph.

    Query must not contain __typename field.

    :param graph: Graph to validate against.
    """

    def __init__(self, graph: Graph):
        self.graph = graph
        self.path = [graph.root]
        self.errors = Errors()

    def visit_field(self, obj: QueryField) -> None:
        node = self.path[-1]
        field = node.fields_map.get(obj.name)
        if field is not None:
            is_field = _AssumeField(node, self.errors).visit(field)
            if is_field:
                for_ = (node.name or "root", obj.name)
                _ValidateOptions(
                    self.graph.data_types, obj.options, for_, self.errors
                ).visit(field)
        else:
            self.errors.report(
                'Field "{}" is not implemented in the "{}" node'.format(
                    obj.name, node.name or "root"
                )
            )

    def visit_link(self, obj: QueryLink) -> None:
        node = self.path[-1]
        graph_obj = node.fields_map.get(obj.name, _undefined)
        if isinstance(graph_obj, Field):
            for_ = (node.name or "root", obj.name)
            _ValidateOptions(
                self.graph.data_types, obj.options, for_, self.errors
            ).visit(graph_obj)

            field_types = _AssumeRecord(self.graph.data_types).visit(
                graph_obj.type
            )
            if field_types is not None:
                fields_validator = _RecordFieldsValidator(
                    self.graph.data_types, field_types, self.errors
                )
                for field in obj.node.fields:
                    fields_validator.visit(field)
            else:
                self.errors.report(
                    'Trying to query "{}.{}" simple field '
                    "as node".format(node.name or "root", obj.name)
                )

        elif isinstance(graph_obj, Link):
            if graph_obj.type_info.type_enum is LinkType.UNION:
                linked_node = self.graph.unions_map[graph_obj.node]
            elif graph_obj.type_info.type_enum is LinkType.INTERFACE:
                linked_node = self.graph.interfaces_map[graph_obj.node]
            else:
                linked_node = self.graph.nodes_map[graph_obj.node]

            for_ = (node.name or "root", obj.name)
            _ValidateOptions(
                self.graph.data_types, obj.options, for_, self.errors
            ).visit(graph_obj)

            self.path.append(linked_node)
            try:
                self.visit(obj.node)
            finally:
                self.path.pop()

        elif graph_obj is _undefined:
            self.errors.report(
                'Link "{}" is not implemented in the "{}" node'.format(
                    obj.name, node.name or "root"
                )
            )
        else:
            raise TypeError(repr(graph_obj))

    def visit_fragment(self, obj: Fragment) -> t.Any:
        graph_node = self.graph.nodes_map[obj.type_name]
        self.path.append(graph_node)
        for field in obj.node.fields:
            if field.name == "__typename":
                continue
            self.visit(field)

        self.path.pop()

    def visit_node(self, obj: QueryNode) -> None:
        fields: t.Dict = {}

        is_union_link = isinstance(self.path[-1], Union)
        is_interface_link = isinstance(self.path[-1], Interface)

        for field in obj.fields:
            if field.name == "__typename":
                continue

            if is_union_link:
                union = self.path[-1]
                if not isinstance(field, Fragment):
                    self.errors.report(
                        "Cannot query field '{}' on type '{}'. "
                        "Did you mean to use an inline fragment on {}?".format(
                            field.name,
                            union.name,
                            " or ".join(
                                [f"'{type_name}'" for type_name in union.types]
                            ),
                        )
                    )
                    continue

                self.visit(field)
                continue
            elif is_interface_link:
                if isinstance(field, Fragment):
                    self.visit(field)
                    continue

                interface = self.path[-1]

                interface_types = self.graph.interfaces_types[interface.name]
                if not interface_types:
                    self.errors.report(
                        "Can not query field '{0}' on interface '{1}'. "
                        "Interface '{1}' is not implemented by any type. "
                        "Add at least one type implementing this interface.".format(  # noqa: E501
                            field.name, interface.name
                        )
                    )
                    continue

                if field.name not in interface.fields_map:
                    implementation = None
                    for impl in interface_types:
                        if field.name in self.graph.nodes_map[impl].fields_map:
                            implementation = impl
                            break

                    self.errors.report(
                        "Can not query field '{}' on type '{}'. "
                        "Did you mean to use an inline fragment on '{}'?".format(  # noqa: E501
                            field.name, interface.name, implementation
                        )
                    )
                    continue

            seen = fields.get(field.result_key)
            if seen is not None:
                if not _field_eq(field, seen):
                    node = self.path[-1].name or "root"
                    self.errors.report(
                        "Found distinct fields with the same "
                        'resulting name "{}" for the node "{}"'.format(
                            field.result_key, node
                        )
                    )
            else:
                fields[field.result_key] = field
            self.visit(field)

            if is_union_link:
                self.path.pop()

        for fr in obj.fragments:
            self.visit(fr)


def validate(graph: Graph, query: QueryNode) -> t.List[str]:
    query_validator = DefaultQueryValidator(graph)
    query_validator.visit(query)
    return query_validator.errors.list



from contextvars import ContextVar
from typing import Iterator, Optional, Type

from prometheus_client.metrics import MetricWrapperBase

from hiku.context import ExecutionContext
from hiku.extensions.base_extension import Extension
from hiku.telemetry.prometheus import (
    AsyncGraphMetrics,
    GraphMetrics,
    GraphMetricsBase,
)


class PrometheusMetrics(Extension):
    def __init__(
        self,
        name: str,
        *,
        metric: Optional[MetricWrapperBase] = None,
        ctx_var: Optional[ContextVar] = None,
        transformer_cls: Type[GraphMetricsBase] = GraphMetrics,
    ):
        self._name = name
        self._metric = metric
        self._ctx_var = ctx_var
        self._transformer = transformer_cls(
            self._name, metric=self._metric, ctx_var=ctx_var
        )

    def on_graph(self, execution_context: ExecutionContext) -> Iterator[None]:
        execution_context.transformers = execution_context.transformers + (
            self._transformer,
        )
        yield

    def on_execute(self, execution_context: ExecutionContext) -> Iterator[None]:
        if self._ctx_var is None:
            yield
        else:
            token = self._ctx_var.set(execution_context.context)
            yield
            self._ctx_var.reset(token)


class PrometheusMetricsAsync(PrometheusMetrics):
    def __init__(
        self,
        name: str,
        *,
        metric: Optional[MetricWrapperBase] = None,
        ctx_var: Optional[ContextVar] = None,
    ):
        super().__init__(
            name,
            metric=metric,
            ctx_var=ctx_var,
            transformer_cls=AsyncGraphMetrics,
        )



import re
import json
import typing as t

from functools import partial
from collections import OrderedDict

from ..directives import (
    Cached,
    Deprecated,
    Directive,
    _SkipDirective,
    _IncludeDirective,
    SchemaDirective,
    get_deprecated,
)
from ..graph import (
    FieldType,
    Graph,
    Root,
    Node,
    Link,
    Option,
    Field,
    Nothing,
    NothingType,
)
from ..graph import GraphVisitor, GraphTransformer
from ..scalar import Scalar
from ..types import (
    EnumRefMeta,
    IDMeta,
    InterfaceRefMeta,
    TypeRef,
    String,
    Sequence,
    Boolean,
    Optional,
    AnyMeta,
    MappingMeta,
    CallableMeta,
    SequenceMeta,
    OptionalMeta,
    TypeRefMeta,
    StringMeta,
    IntegerMeta,
    FloatMeta,
    BooleanMeta,
    UnionRefMeta,
)
from ..types import Any, RecordMeta, AbstractTypeVisitor
from ..utils import (
    listify,
    cached_property,
)
from .types import (
    ENUM,
    EnumValueIdent,
    INTERFACE,
    SCALAR,
    NON_NULL,
    LIST,
    INPUT_OBJECT,
    OBJECT,
    UNION,
    DIRECTIVE,
    FieldIdent,
    FieldArgIdent,
    InputObjectFieldIdent,
    DirectiveArgIdent,
    HashedNamedTuple,
)
from ..utils.serialize import serialize

_BUILTIN_DIRECTIVES: t.Tuple[
    t.Union[t.Type[Directive], t.Type[SchemaDirective]], ...
] = (
    _SkipDirective,
    _IncludeDirective,
    Deprecated,
    Cached,
)

BUILTIN_SCALARS: t.Tuple[SCALAR, ...] = (  # type: ignore[valid-type]
    SCALAR("String"),
    SCALAR("Int"),
    SCALAR("Boolean"),
    SCALAR("Float"),
    SCALAR("Any"),
    SCALAR("ID"),
)


def _async_wrapper(func: t.Callable) -> t.Callable:
    async def wrapper(*args: t.Any, **kwargs: t.Any) -> t.Any:
        return func(*args, **kwargs)

    return wrapper


QUERY_ROOT_NAME = "Query"
MUTATION_ROOT_NAME = "Mutation"


class SchemaInfo:
    def __init__(
        self,
        query_graph: Graph,
        mutation_graph: t.Optional[Graph],
    ):
        self.query_graph = query_graph
        self.data_types = query_graph.data_types
        self.mutation_graph = mutation_graph
        self.directives = _BUILTIN_DIRECTIVES + tuple(
            query_graph.directives or ()
        )
        self.nodes_map = self._nodes_map()

    def _nodes_map(self) -> OrderedDict:
        nodes = [(n.name, n) for n in self.query_graph.nodes]
        nodes.append((QUERY_ROOT_NAME, self.query_graph.root))
        if self.mutation_graph is not None:
            nodes.append((MUTATION_ROOT_NAME, self.mutation_graph.root))
        return OrderedDict(nodes)

    @cached_property
    def directives_map(self) -> OrderedDict:
        return OrderedDict(
            (d.__directive_info__.name, d) for d in self.directives
        )

    @staticmethod
    def is_field_hidden(field: Field) -> bool:
        """Determines if a field should be hidden from introspection."""
        return field.name.startswith("_")

    @staticmethod
    def is_data_type_as_input(name: str) -> bool:
        return True


class TypeIdent(AbstractTypeVisitor):
    def __init__(self, graph: Graph, input_mode: bool = False) -> None:
        self._graph = graph
        self._input_mode = input_mode

    def visit_any(self, obj: AnyMeta) -> HashedNamedTuple:
        return SCALAR("Any")

    def visit_scalar(self, obj: t.Type[Scalar]) -> HashedNamedTuple:
        return NON_NULL(SCALAR(obj.__type_name__))

    def visit_mapping(self, obj: MappingMeta) -> HashedNamedTuple:
        return SCALAR("Any")

    def visit_record(self, obj: RecordMeta) -> HashedNamedTuple:
        return SCALAR("Any")

    def visit_callable(self, obj: CallableMeta) -> t.NoReturn:
        raise TypeError("Not expected here: {!r}".format(obj))

    def visit_sequence(self, obj: SequenceMeta) -> HashedNamedTuple:
        return NON_NULL(LIST(self.visit(obj.__item_type__)))

    def visit_optional(self, obj: OptionalMeta) -> HashedNamedTuple:
        ident = self.visit(obj.__type__)
        return ident.of_type if isinstance(ident, NON_NULL) else ident

    def visit_typeref(self, obj: TypeRefMeta) -> HashedNamedTuple:
        if self._input_mode:
            assert (
                obj.__type_name__ in self._graph.data_types
            ), obj.__type_name__
            return NON_NULL(INPUT_OBJECT(obj.__type_name__))
        else:
            return NON_NULL(OBJECT(obj.__type_name__))

    def visit_unionref(self, obj: UnionRefMeta) -> t.Any:
        return NON_NULL(UNION(obj.__type_name__, tuple()))

    def visit_interfaceref(self, obj: InterfaceRefMeta) -> t.Any:
        return NON_NULL(INTERFACE(obj.__type_name__, tuple()))

    def visit_enumref(self, obj: EnumRefMeta) -> t.Any:
        return NON_NULL(
            ENUM(
                obj.__type_name__,
                tuple(),
            )
        )

    def visit_string(self, obj: StringMeta) -> HashedNamedTuple:
        return NON_NULL(SCALAR("String"))

    def visit_id(self, obj: IDMeta) -> t.Any:
        return NON_NULL(SCALAR("ID"))

    def visit_integer(self, obj: IntegerMeta) -> HashedNamedTuple:
        return NON_NULL(SCALAR("Int"))

    def visit_float(self, obj: FloatMeta) -> HashedNamedTuple:
        return NON_NULL(SCALAR("Float"))

    def visit_boolean(self, obj: BooleanMeta) -> HashedNamedTuple:
        return NON_NULL(SCALAR("Boolean"))


def na_maybe(schema: SchemaInfo) -> NothingType:
    return Nothing


def schema_link(schema: SchemaInfo) -> None:
    return None


def type_link(
    schema: SchemaInfo, options: t.Dict
) -> t.Union[HashedNamedTuple, NothingType]:
    name = options["name"]
    if name in schema.nodes_map:
        return OBJECT(name)
    elif name in schema.query_graph.unions_map:
        union = schema.query_graph.unions_map[name]
        return UNION(
            union.name, tuple(OBJECT(type_name) for type_name in union.types)
        )
    elif name in schema.query_graph.interfaces_map:
        interface = schema.query_graph.interfaces_map[name]
        possible_types = schema.query_graph.interfaces_types[name]
        return INTERFACE(
            interface.name,
            tuple(OBJECT(type_name) for type_name in possible_types),
        )
    elif name in schema.query_graph.enums_map:
        enum = schema.query_graph.enums_map[name]
        return ENUM(
            enum.name,
            tuple(OBJECT(value.name) for value in enum.values),
        )
    elif name in schema.query_graph.scalars_map:
        return SCALAR(name)
    else:
        return Nothing


@listify
def root_schema_types(schema: SchemaInfo) -> t.Iterator[HashedNamedTuple]:
    for scalar in BUILTIN_SCALARS:
        yield scalar

    for name in schema.nodes_map:
        yield OBJECT(name)

    for name, type_ in schema.data_types.items():
        if isinstance(type_, RecordMeta):
            yield OBJECT(name)
            if schema.is_data_type_as_input(name):
                yield INPUT_OBJECT(name)

    for union in schema.query_graph.unions:
        yield UNION(
            union.name, tuple(OBJECT(type_name) for type_name in union.types)
        )

    for interface in schema.query_graph.interfaces:
        yield INTERFACE(
            interface.name,
            tuple(
                OBJECT(type_name)
                for type_name in schema.query_graph.interfaces_types[
                    interface.name
                ]
            ),
        )

    for enum in schema.query_graph.enums:
        yield ENUM(
            enum.name,
            tuple(OBJECT(value.name) for value in enum.values),
        )

    # custom scalars
    for scalar in schema.query_graph.scalars:
        yield SCALAR(scalar.__type_name__)


def root_schema_query_type(schema: SchemaInfo) -> HashedNamedTuple:
    return OBJECT(QUERY_ROOT_NAME)


def root_schema_mutation_type(
    schema: SchemaInfo,
) -> t.Union[HashedNamedTuple, NothingType]:
    if schema.mutation_graph is not None:
        return OBJECT(MUTATION_ROOT_NAME)
    else:
        return Nothing


def root_schema_directives(schema: SchemaInfo) -> t.List[DIRECTIVE]:  # type: ignore[valid-type]  # noqa: E501
    return [
        DIRECTIVE(directive.__directive_info__.name)
        for directive in schema.directives
    ]


@listify
def type_info(
    schema: SchemaInfo, fields: t.List[Field], ids: t.List
) -> t.Iterator[t.List[t.Optional[t.Dict]]]:
    for ident in ids:
        if isinstance(ident, OBJECT):
            if ident.name in schema.nodes_map:
                description = schema.nodes_map[ident.name].description
            else:
                description = None
            info = {
                "id": ident,
                "kind": "OBJECT",
                "name": ident.name,
                "description": description,
            }
        elif isinstance(ident, INPUT_OBJECT):
            info = {
                "id": ident,
                "kind": "INPUT_OBJECT",
                "name": "IO{}".format(ident.name),
                "description": None,
            }
        elif isinstance(ident, NON_NULL):
            info = {"id": ident, "kind": "NON_NULL"}
        elif isinstance(ident, LIST):
            info = {"id": ident, "kind": "LIST"}
        elif isinstance(ident, SCALAR):
            info = {"id": ident, "name": ident.name, "kind": "SCALAR"}
        elif isinstance(ident, UNION):
            info = {
                "id": ident,
                "kind": "UNION",
                "name": ident.name,
                "description": schema.query_graph.unions_map[
                    ident.name
                ].description,
            }
        elif isinstance(ident, INTERFACE):
            info = {
                "id": ident,
                "kind": "INTERFACE",
                "name": ident.name,
                "description": schema.query_graph.interfaces_map[
                    ident.name
                ].description,
            }
        elif isinstance(ident, ENUM):
            info = {
                "id": ident,
                "kind": "ENUM",
                "name": ident.name,
                "description": schema.query_graph.enums_map[
                    ident.name
                ].description,
            }
        else:
            raise TypeError(repr(ident))
        yield [info.get(f.name) for f in fields]


@listify
def type_fields_link(
    schema: SchemaInfo, ids: t.List, options: t.Dict
) -> t.Iterator[t.List[HashedNamedTuple]]:
    for ident in ids:
        if isinstance(ident, OBJECT):
            if ident.name in schema.nodes_map:
                node = schema.nodes_map[ident.name]
                field_idents = [
                    FieldIdent(ident.name, f.name)
                    for f in node.fields
                    if not schema.is_field_hidden(f)
                ]
            else:
                type_ = schema.data_types[ident.name]
                field_idents = [
                    FieldIdent(ident.name, f_name)
                    for f_name, f_type in type_.__field_types__.items()
                ]
            if not field_idents:
                raise TypeError(
                    'Object type "{}" does not contain fields, '
                    "which is not acceptable for GraphQL in order "
                    "to define schema type".format(ident.name)
                )
            yield field_idents
        elif isinstance(ident, INTERFACE):
            interface = schema.query_graph.interfaces_map[ident.name]
            yield [
                FieldIdent(ident.name, f.name)
                for f in interface.fields
                if not schema.is_field_hidden(f)
            ]
        else:
            yield []


@listify
def type_of_type_link(
    schema: SchemaInfo, ids: t.List
) -> t.Iterator[t.Union[HashedNamedTuple, NothingType]]:
    for ident in ids:
        if isinstance(ident, (NON_NULL, LIST)):
            yield ident.of_type
        else:
            yield Nothing


@listify
def possible_types_type_link(schema: SchemaInfo, ids: t.List) -> t.Iterator:
    if ids is None:
        yield []

    for ident in ids:
        if isinstance(ident, (UNION, INTERFACE)):
            yield ident.possible_types
        else:
            yield []


@listify
def enum_values_type_link(
    schema: SchemaInfo, ids: t.List, opts: t.Dict
) -> t.Iterator:
    if ids is None:
        yield []

    for ident in ids:
        if isinstance(ident, ENUM):
            yield [
                EnumValueIdent(ident.name, value.name)
                for value in ident.of_types
            ]
        else:
            yield []


@listify
def interfaces_type_link(schema: SchemaInfo, ids: t.List) -> t.Iterator:
    if ids is None:
        yield []

    for ident in ids:
        if isinstance(ident, OBJECT) and ident.name in schema.nodes_map:
            node = schema.nodes_map[ident.name]
            yield [
                INTERFACE(interface, tuple()) for interface in node.implements
            ]
        else:
            yield []


@listify
def field_info(
    schema: SchemaInfo, fields: t.List[Field], ids: t.List
) -> t.Iterator[t.List[t.Dict]]:
    for ident in ids:
        if ident.node in schema.nodes_map:
            node = schema.nodes_map[ident.node]
            field = node.fields_map[ident.name]
            deprecated = None
            if isinstance(field, (Field, Link)):
                deprecated = get_deprecated(field)

            info = {
                "id": ident,
                "name": field.name,
                "description": field.description,
                "isDeprecated": bool(deprecated),
                "deprecationReason": deprecated and deprecated.reason,
            }
        else:
            info = {
                "id": ident,
                "name": ident.name,
                "description": None,
                "isDeprecated": False,
                "deprecationReason": None,
            }
        yield [info[f.name] for f in fields]


@listify
def field_type_link(
    schema: SchemaInfo, ids: t.List
) -> t.Iterator[HashedNamedTuple]:
    type_ident = TypeIdent(schema.query_graph)
    for ident in ids:
        if ident.node in schema.nodes_map:
            node = schema.nodes_map[ident.node]
            field = node.fields_map[ident.name]
            yield type_ident.visit(field.type or Any)
        elif ident.node in schema.query_graph.interfaces_map:
            interface = schema.query_graph.interfaces_map[ident.node]
            field = interface.fields_map[ident.name]
            yield type_ident.visit(field.type or Any)
        else:
            data_type = schema.data_types[ident.node]
            field_type = data_type.__field_types__[ident.name]
            yield type_ident.visit(field_type)


@listify
def field_args_link(
    schema: SchemaInfo, ids: t.List
) -> t.Iterator[t.List[HashedNamedTuple]]:
    for ident in ids:
        if ident.node in schema.nodes_map:
            node = schema.nodes_map[ident.node]
            field = node.fields_map[ident.name]
            yield [
                FieldArgIdent(ident.node, field.name, option.name)
                for option in field.options
            ]
        else:
            yield []


@listify
def type_input_object_input_fields_link(
    schema: SchemaInfo, ids: t.List
) -> t.Iterator[t.List[HashedNamedTuple]]:
    for ident in ids:
        if isinstance(ident, INPUT_OBJECT):
            data_type = schema.data_types[ident.name]
            yield [
                InputObjectFieldIdent(ident.name, key)
                for key in data_type.__field_types__.keys()
            ]
        else:
            yield []


@listify
def input_value_info(
    schema: SchemaInfo, fields: t.List[Field], ids: t.List[HashedNamedTuple]
) -> t.Iterator[t.List[t.Dict]]:
    for ident in ids:
        if isinstance(ident, FieldArgIdent):
            node = schema.nodes_map[ident.node]
            field = node.fields_map[ident.field]
            option = field.options_map[ident.name]
            if option.default is Nothing:
                default = None
            else:
                if (
                    option.type_info
                    and option.type_info.type_enum is FieldType.ENUM
                ):
                    enum = schema.query_graph.enums_map[
                        option.type_info.type_name
                    ]
                    default = serialize(
                        option.type, option.default, enum.serialize
                    )
                elif (
                    option.type_info
                    and option.type_info.type_enum is FieldType.CUSTOM_SCALAR
                ):
                    scalar = schema.query_graph.scalars_map[
                        option.type_info.type_name
                    ]
                    default = json.dumps(
                        serialize(option.type, option.default, scalar.serialize)
                    )
                else:
                    default = json.dumps(option.default)
            info = {
                "id": ident,
                "name": option.name,
                "description": option.description,
                "defaultValue": default,
            }
            yield [info[f.name] for f in fields]
        elif isinstance(ident, InputObjectFieldIdent):
            info = {
                "id": ident,
                "name": ident.key,
                "description": None,
                "defaultValue": None,
            }
            yield [info[f.name] for f in fields]
        elif isinstance(ident, DirectiveArgIdent):
            directive = schema.directives_map[ident.name]
            arg = directive.args_map()[ident.arg]
            info = {
                "id": ident,
                "name": arg.field_name,
                "description": arg.description,
                "defaultValue": arg.default_value,
            }
            yield [info[f.name] for f in fields]
        else:
            raise TypeError(repr(ident))


@listify
def input_value_type_link(
    schema: SchemaInfo, ids: t.List[HashedNamedTuple]
) -> t.Iterator[HashedNamedTuple]:
    type_ident = TypeIdent(schema.query_graph, input_mode=True)
    for ident in ids:
        if isinstance(ident, FieldArgIdent):
            node = schema.nodes_map[ident.node]
            field = node.fields_map[ident.field]
            option = field.options_map[ident.name]
            yield type_ident.visit(option.type)
        elif isinstance(ident, InputObjectFieldIdent):
            data_type = schema.data_types[ident.name]
            field_type = data_type.__field_types__[ident.key]
            yield type_ident.visit(field_type)
        elif isinstance(ident, DirectiveArgIdent):
            directive = schema.directives_map[ident.name]
            arg = directive.args_map()[ident.arg]
            yield arg.type_ident
        else:
            raise TypeError(repr(ident))


@listify
def directive_value_info(
    schema: SchemaInfo,
    fields: t.List[Field],
    ids: t.List[DIRECTIVE],  # type: ignore[valid-type]
) -> t.Iterator[t.List[Any]]:
    for ident in ids:
        if ident.name in schema.directives_map:  # type: ignore[attr-defined]
            info = schema.directives_map[ident.name].__directive_info__  # type: ignore  # noqa: E501
            data = {
                "name": info.name,
                "description": info.description,
                "locations": [loc.value for loc in info.locations],
            }
            yield [data[f.name] for f in fields]


def directive_args_link(
    schema: SchemaInfo, ids: t.List[str]
) -> t.List[t.List[DirectiveArgIdent]]:  # type: ignore[valid-type]
    links = []
    for ident in ids:
        directive = schema.directives_map[ident].__directive_info__
        links.append(
            [DirectiveArgIdent(ident, arg.name) for arg in directive.args]
        )
    return links


@listify
def enum_value_info(
    schema: SchemaInfo,
    fields: t.List[Field],
    ids: t.List[EnumValueIdent],  # type: ignore[valid-type]
) -> t.Iterator[t.List[Any]]:
    for ident in ids:
        enum = schema.query_graph.enums_map[ident.enum_name]  # type: ignore[attr-defined]  # noqa: E501
        value = enum.values_map[ident.value_name]  # type: ignore[attr-defined]
        data = {
            "name": value.name,
            "description": value.description,
            "isDeprecated": bool(value.deprecation_reason),
            "deprecationReason": value.deprecation_reason,
        }
        yield [data[f.name] for f in fields]


GRAPH = Graph(
    [
        Node(
            "__Type",
            [
                Field("id", None, type_info),
                Field("kind", String, type_info),
                Field("name", String, type_info),
                Field("description", String, type_info),
                # OBJECT and INTERFACE only
                Link(
                    "fields",
                    Sequence[TypeRef["__Field"]],
                    type_fields_link,
                    requires="id",
                    options=[
                        Option("includeDeprecated", Boolean, default=False)
                    ],
                ),
                # OBJECT only
                Link(
                    "interfaces",
                    Sequence[TypeRef["__Type"]],
                    interfaces_type_link,
                    requires="id",
                ),
                # INTERFACE and UNION only
                Link(
                    "possibleTypes",
                    Sequence[TypeRef["__Type"]],
                    possible_types_type_link,
                    requires="id",
                ),
                # ENUM only
                Link(
                    "enumValues",
                    Sequence[TypeRef["__EnumValue"]],
                    enum_values_type_link,
                    requires="id",
                    options=[
                        Option("includeDeprecated", Boolean, default=False)
                    ],
                ),
                # INPUT_OBJECT only
                Link(
                    "inputFields",
                    Sequence[TypeRef["__InputValue"]],
                    type_input_object_input_fields_link,
                    requires="id",
                ),
                # NON_NULL and LIST only
                Link(
                    "ofType",
                    Optional[TypeRef["__Type"]],
                    type_of_type_link,
                    requires="id",
                ),
            ],
        ),
        Node(
            "__Field",
            [
                Field("id", None, field_info),
                Field("name", String, field_info),
                Field("description", String, field_info),
                Link(
                    "args",
                    Sequence[TypeRef["__InputValue"]],
                    field_args_link,
                    requires="id",
                ),
                Link("type", TypeRef["__Type"], field_type_link, requires="id"),
                Field("isDeprecated", Boolean, field_info),
                Field("deprecationReason", String, field_info),
            ],
        ),
        Node(
            "__InputValue",
            [
                Field("id", None, input_value_info),
                Field("name", String, input_value_info),
                Field("description", String, input_value_info),
                Link(
                    "type",
                    TypeRef["__Type"],
                    input_value_type_link,
                    requires="id",
                ),
                Field("defaultValue", String, input_value_info),
            ],
        ),
        Node(
            "__Directive",
            [
                Field("name", String, directive_value_info),
                Field("description", String, directive_value_info),
                Field("locations", Sequence[String], directive_value_info),
                Link(
                    "args",
                    Sequence[TypeRef["__InputValue"]],
                    directive_args_link,
                    requires="name",
                ),
            ],
        ),
        Node(
            "__EnumValue",
            [
                Field("name", String, enum_value_info),
                Field("description", String, enum_value_info),
                Field("isDeprecated", Boolean, enum_value_info),
                Field("deprecationReason", String, enum_value_info),
            ],
        ),
        Node(
            "__Schema",
            [
                Link(
                    "types",
                    Sequence[TypeRef["__Type"]],
                    root_schema_types,
                    requires=None,
                ),
                Link(
                    "queryType",
                    TypeRef["__Type"],
                    root_schema_query_type,
                    requires=None,
                ),
                Link(
                    "mutationType",
                    Optional[TypeRef["__Type"]],
                    root_schema_mutation_type,
                    requires=None,
                ),
                Link(
                    "subscriptionType",
                    Optional[TypeRef["__Type"]],
                    na_maybe,
                    requires=None,
                ),
                Link(
                    "directives",
                    Sequence[TypeRef["__Directive"]],
                    root_schema_directives,
                    requires=None,
                ),
            ],
        ),
        Root(
            [
                Link(
                    "__schema", TypeRef["__Schema"], schema_link, requires=None
                ),
                Link(
                    "__type",
                    Optional[TypeRef["__Type"]],
                    type_link,
                    requires=None,
                    options=[Option("name", String)],
                ),
            ]
        ),
    ]
)


class ValidateGraph(GraphVisitor):
    _name_re = re.compile(r"^[_a-zA-Z]\w*$", re.ASCII)

    def __init__(self) -> None:
        self._path: t.List[str] = []
        self._errors: t.List[str] = []

    def _add_error(self, name: str, description: str) -> None:
        path = ".".join(self._path + [name])
        self._errors.append("{}: {}".format(path, description))

    @classmethod
    def validate(cls, graph: Graph) -> None:
        self = cls()
        self.visit(graph)
        if self._errors:
            raise ValueError(
                "Invalid GraphQL graph:\n{}".format(
                    "\n".join("- {}".format(err) for err in self._errors)
                )
            )

    def visit_node(self, obj: Node) -> None:
        assert obj.name is not None
        if not self._name_re.match(obj.name):
            self._add_error(obj.name, "Invalid node name: {}".format(obj.name))
        if obj.fields:
            self._path.append(obj.name)
            super(ValidateGraph, self).visit_node(obj)
            self._path.pop()
        else:
            self._add_error(
                obj.name, "No fields in the {} node".format(obj.name)
            )

    def visit_root(self, obj: Root) -> None:
        if obj.fields:
            self._path.append("Root")
            super(ValidateGraph, self).visit_root(obj)
            self._path.pop()
        else:
            self._add_error("Root", "No fields in the Root node")

    def visit_field(self, obj: Field) -> None:
        if not self._name_re.match(obj.name):
            self._add_error(obj.name, "Invalid field name: {}".format(obj.name))
        super(ValidateGraph, self).visit_field(obj)

    def visit_link(self, obj: Link) -> None:
        if not self._name_re.match(obj.name):
            self._add_error(obj.name, "Invalid link name: {}".format(obj.name))
        super(ValidateGraph, self).visit_link(obj)

    def visit_option(self, obj: Option) -> None:
        if not self._name_re.match(obj.name):
            self._add_error(
                obj.name, "Invalid option name: {}".format(obj.name)
            )

        super(ValidateGraph, self).visit_option(obj)


class BindToSchema(GraphTransformer):
    def __init__(self, schema: SchemaInfo) -> None:
        self.schema = schema
        self._processed: t.Dict = {}

    def visit_field(self, obj: Field) -> Field:
        field = super(BindToSchema, self).visit_field(obj)
        func = self._processed.get(obj.func)
        if func is None:
            func = self._processed[obj.func] = partial(obj.func, self.schema)
        field.func = func
        return field

    def visit_link(self, obj: Link) -> Link:
        link = super(BindToSchema, self).visit_link(obj)
        link.func = partial(link.func, self.schema)
        return link


class MakeAsync(GraphTransformer):
    def __init__(self) -> None:
        self._processed: t.Dict = {}

    def visit_field(self, obj: Field) -> Field:
        field = super(MakeAsync, self).visit_field(obj)
        func = self._processed.get(obj.func)
        if func is None:
            func = self._processed[obj.func] = _async_wrapper(obj.func)
        field.func = func
        return field

    def visit_link(self, obj: Link) -> Link:
        link = super(MakeAsync, self).visit_link(obj)
        link.func = _async_wrapper(link.func)
        return link


def type_name_field_func(
    node_name: str, fields: t.List[Field], ids: t.Optional[t.List] = None
) -> t.List:
    return [[node_name] for _ in ids] if ids is not None else [node_name]


class AddIntrospection(GraphTransformer):
    def __init__(
        self, introspection_graph: Graph, type_name_field_factory: t.Callable
    ):
        self.introspection_graph = introspection_graph
        self.type_name_field_factory = type_name_field_factory

    def visit_node(self, obj: Node) -> Node:
        node = super(AddIntrospection, self).visit_node(obj)
        node.fields.append(self.type_name_field_factory(obj.name))
        return node

    def visit_root(self, obj: Root) -> Root:
        root = super(AddIntrospection, self).visit_root(obj)
        root.fields.append(self.type_name_field_factory(QUERY_ROOT_NAME))
        return root

    def visit_graph(self, obj: Graph) -> Graph:
        graph = super(AddIntrospection, self).visit_graph(obj)
        graph.items.extend(self.introspection_graph.items)
        return graph


class GraphQLIntrospection(GraphTransformer):
    """Adds GraphQL introspection into synchronous graph

    Example:

    .. code-block:: python

        from hiku.graph import apply
        from hiku.introspection.graphql import GraphQLIntrospection

        graph = apply(graph, [GraphQLIntrospection(graph)])

    """

    def __init__(
        self,
        query_graph: Graph,
        mutation_graph: t.Optional[Graph] = None,
    ) -> None:
        """
        :param query_graph: graph, where Root node represents Query root
            operation type
        :param mutation_graph: graph, where Root node represents Mutation root
            operation type
        """
        self.schema = SchemaInfo(
            query_graph,
            mutation_graph,
        )

    def __type_name__(self, node_name: t.Optional[str]) -> Field:
        return Field(
            "__typename", String, partial(type_name_field_func, node_name)
        )

    def __introspection_graph__(self) -> Graph:
        return BindToSchema(self.schema).visit(GRAPH)

    def visit_node(self, obj: Node) -> Node:
        node = super(GraphQLIntrospection, self).visit_node(obj)
        node.fields.append(self.__type_name__(obj.name))
        return node

    def visit_root(self, obj: Root) -> Root:
        root = super(GraphQLIntrospection, self).visit_root(obj)
        root.fields.append(self.__type_name__(QUERY_ROOT_NAME))
        return root

    def visit_graph(self, obj: Graph) -> Graph:
        ValidateGraph.validate(obj)
        introspection_graph = self.__introspection_graph__()
        items = [self.visit(node) for node in obj.items]
        items.extend(introspection_graph.items)
        return Graph(
            items,
            data_types=obj.data_types,
            directives=obj.directives,
            unions=obj.unions,
            interfaces=obj.interfaces,
            enums=obj.enums,
            scalars=obj.scalars,
        )


class AsyncGraphQLIntrospection(GraphQLIntrospection):
    """Adds GraphQL introspection into asynchronous graph

    Example:

    .. code-block:: python

        from hiku.graph import apply
        from hiku.introspection.graphql import AsyncGraphQLIntrospection

        graph = apply(graph, [AsyncGraphQLIntrospection(graph)])

    """

    def __type_name__(self, node_name: t.Optional[str]) -> Field:
        return Field(
            "__typename",
            String,
            _async_wrapper(partial(type_name_field_func, node_name)),
        )

    def __introspection_graph__(self) -> Graph:
        graph = super(AsyncGraphQLIntrospection, self).__introspection_graph__()
        graph = MakeAsync().visit(graph)
        return graph



from typing import (
    Callable,
    Optional,
)

from hiku.graph import Field, Graph
from hiku.introspection.graphql import GraphQLIntrospection
from hiku.introspection.graphql import AsyncGraphQLIntrospection


def _is_field_hidden_wrapper(func: Callable) -> Callable:
    def wrapper(field: Field) -> bool:
        if field.name in ("_entities", "_service"):
            return False

        return func(field)

    return wrapper


def _is_data_type_as_input_wrapper(func: Callable) -> Callable:
    def wrapper(name: str) -> bool:
        if name in ("_Service",):
            return False

        return func(name)

    return wrapper


class BaseFederatedGraphQLIntrospection(GraphQLIntrospection):
    def __init__(
        self,
        query_graph: Graph,
        mutation_graph: Optional[Graph] = None,
    ) -> None:
        super().__init__(query_graph, mutation_graph)
        self.schema.is_field_hidden = _is_field_hidden_wrapper(  # type: ignore
            self.schema.is_field_hidden
        )
        self.schema.is_data_type_as_input = _is_data_type_as_input_wrapper(  # type: ignore  # noqa: E501
            self.schema.is_data_type_as_input
        )


class FederatedGraphQLIntrospection(BaseFederatedGraphQLIntrospection):
    """
    Federation-aware introspection for Federation
    https://www.apollographql.com/docs/federation/federation-spec/#federation-schema-specification
    """


class AsyncFederatedGraphQLIntrospection(
    FederatedGraphQLIntrospection, AsyncGraphQLIntrospection
):
    """Adds GraphQL introspection into asynchronous federated graph"""



from typing import Callable, Dict, Iterator

from hiku.context import ExecutionContext
from hiku.extensions.base_extension import Extension


class CustomContext(Extension):
    def __init__(
        self,
        get_context: Callable[[ExecutionContext], Dict],
    ):
        self.get_context = get_context

    def on_execute(self, execution_context: ExecutionContext) -> Iterator[None]:
        execution_context.context = self.get_context(execution_context)
        yield



import abc
import dataclasses
import enum

from typing import (
    Any,
    Generic,
    Optional,
    Sequence,
    TYPE_CHECKING,
    TypeVar,
    Union,
)

if TYPE_CHECKING:
    from hiku.graph import AbstractGraphVisitor


@dataclasses.dataclass
class EnumValue:
    name: str
    description: Optional[str] = None
    deprecation_reason: Optional[str] = None


class BaseEnum(abc.ABC):
    def __init__(
        self,
        name: str,
        values: Sequence[Union[str, EnumValue]],
        description: Optional[str] = None,
    ):
        self.name = name
        self.description = description
        self.values = [
            EnumValue(v) if isinstance(v, str) else v for v in values
        ]
        self.values_map = {v.name: v for v in self.values}

    @abc.abstractmethod
    def parse(self, value: Any) -> Any:
        raise NotImplementedError

    @abc.abstractmethod
    def serialize(self, value: Any) -> str:
        raise NotImplementedError

    def __contains__(self, item: str) -> bool:
        return item in self.values_map

    def accept(self, visitor: "AbstractGraphVisitor") -> Any:
        return visitor.visit_enum(self)


EM = TypeVar("EM", bound=enum.EnumMeta)
E = TypeVar("E", bound=enum.Enum)


class Enum(BaseEnum):
    def parse(self, value: Any) -> str:
        if value not in self:
            raise TypeError(
                "Enum '{}' can not represent value: {!r}".format(
                    self.name, value
                )
            )
        return value

    def serialize(self, value: str) -> str:
        if value not in self:
            raise TypeError(
                "Enum '{}' can not represent value: {!r}".format(
                    self.name, value
                )
            )
        return value

    @classmethod
    def from_builtin(
        cls,
        enum_cls: EM,
        name: Optional[str] = None,
        description: Optional[str] = None,
    ) -> "EnumFromBuiltin":
        return EnumFromBuiltin(enum_cls, name or enum_cls.__name__, description)


class EnumFromBuiltin(BaseEnum, Generic[E]):
    def __init__(
        self, enum_cls: EM, name: str, description: Optional[str] = None
    ):
        super().__init__(name, [v for v in enum_cls.__members__], description)
        self.enum_cls = enum_cls

    def parse(self, value: Any) -> E:
        try:
            return self.enum_cls(value)
        except ValueError:
            raise TypeError(
                "Enum '{}' can not represent value: {!r}".format(
                    self.name, value
                )
            )

    def serialize(self, value: Any) -> str:
        if not isinstance(value, self.enum_cls):
            raise TypeError(
                "Enum '{}' can not represent value: {!r}".format(
                    self.name, value
                )
            )

        return value.name



from functools import wraps
from typing import (
    Any,
    Callable,
    TYPE_CHECKING,
    TypeVar,
    List,
    Iterator,
)


from hiku.compat import ParamSpec

from .immutable import ImmutableDict, to_immutable_dict
from .cached_property import cached_property
from .const import const, Const

if TYPE_CHECKING:
    from hiku.query import Field


T = TypeVar("T")
P = ParamSpec("P")


def listify(func: Callable[P, Iterator[T]]) -> Callable[P, List[T]]:
    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> List[T]:
        return list(func(*args, **kwargs))

    return wrapper


def empty_field(fields: List["Field"], ids: Any) -> Any:
    return [[None]] * len(ids)


__all__ = [
    "ImmutableDict",
    "to_immutable_dict",
    "cached_property",
    "const",
    "Const",
]
